from contextlib import nullcontext as does_not_raise

import jax.numpy as jnp
import pytest

from nemos import validation


@pytest.mark.parametrize(
    "tree, expectation",
    [
        (jnp.array([[1], [2], [3]]), does_not_raise()),
        (
            jnp.array([[1], [2], [jnp.nan]]),
            pytest.raises(ValueError, match="The provided trees contain Nans"),
        ),
        (
            jnp.array([[1], [jnp.inf], [3]]),
            pytest.raises(ValueError, match="The provided trees contain Infs"),
        ),
        (
            jnp.array([[1], [jnp.inf], [jnp.nan]]),
            pytest.raises(ValueError, match="The provided trees contain Infs and Nans"),
        ),
    ],
)
def test_error_invalid_entry(tree, expectation):
    """Test validation of trees generates the correct exceptions."""
    valid_data = jnp.array([[1], [2], [3]])
    with expectation:
        validation.error_invalid_entry(valid_data, tree)


@pytest.mark.parametrize(
    "fill_val, expectation",
    [
        (0, does_not_raise()),
        (
            jnp.inf,
            pytest.raises(
                ValueError, match="At least a NaN or an Inf at all sample points"
            ),
        ),
        (
            jnp.nan,
            pytest.raises(
                ValueError, match="At least a NaN or an Inf at all sample points"
            ),
        ),
    ],
)
def test_error_all_invalid(fill_val, expectation):
    """Test error when all samples have at least a nan."""
    inp = jnp.array(
        [[fill_val, 0, 0], [0, 0, fill_val], [0, 0, fill_val], [0, fill_val, 0]]
    )
    valid_inp = jnp.zeros((inp.shape[0], 1))
    # some structured tree
    trees = [[[valid_inp], inp], valid_inp]
    with expectation:
        validation.error_all_invalid(*trees)


# Sample pytree for testing
sample_pytree = {"a": jnp.array([1, 2, 3]), "b": {"c": jnp.array([[1], [2]])}}

sample_pytree_same_shape = {
    "a": jnp.array([[1, 2, 3]]),
    "b": {"c": jnp.array([[1, 2]])},
}


@pytest.mark.parametrize(
    "x, expected_len, err_message, expectation",
    [
        ([1, 2], 2, "", does_not_raise()),
        ([1, 2, 3], 2, "Length does not match expected.", pytest.raises(ValueError)),
    ],
)
def test_check_length(x, expected_len, err_message, expectation):
    with expectation:
        validation.check_length(x, expected_len, err_message)


@pytest.mark.parametrize(
    "tree, err_message, data_type, expectation",
    [
        (sample_pytree, "", jnp.float32, does_not_raise()),
        (
            [1, 2, "invalid"],
            "Conversion failed.",
            jnp.float32,
            pytest.raises(TypeError),
        ),
    ],
)
def test_convert_tree_leaves_to_jax_array(tree, err_message, data_type, expectation):
    with expectation:
        validation.convert_tree_leaves_to_jax_array(tree, err_message, data_type)


@pytest.mark.parametrize(
    "tree, expected_dim, err_message, expectation",
    [
        (sample_pytree_same_shape, 2, "", does_not_raise()),
        (
            {"a": jnp.array([[1], [2]]), "b": jnp.array([1, 2, 3])},
            1,
            "Dimensionality mismatch.",
            pytest.raises(ValueError),
        ),
    ],
)
def test_check_tree_leaves_dimensionality(tree, expected_dim, err_message, expectation):
    with expectation:
        validation.check_tree_leaves_dimensionality(tree, expected_dim, err_message)


@pytest.mark.parametrize(
    "arrays, axis, err_message, expectation",
    [
        ((jnp.array([1, 2]), jnp.array([3, 4])), 0, "", does_not_raise()),
        (
            (jnp.array([1, 2]), jnp.array([[3, 4, 3], [5, 6, 1]])),
            0,
            "",
            does_not_raise(),
        ),
        (
            (jnp.array([1, 2, 3]), jnp.array([[3, 4, 3], [5, 6, 1]])),
            0,
            "Shape mismatch on axis.",
            pytest.raises(ValueError),
        ),
    ],
)
def test_check_same_shape_on_axis(arrays, axis, err_message, expectation):
    with expectation:
        validation.check_same_shape_on_axis(*arrays, axis=axis, err_message=err_message)


@pytest.mark.parametrize(
    "tree, array, axis, err_message, expectation",
    [
        (sample_pytree_same_shape, jnp.array([[1, 2, 3]]), 0, "", does_not_raise()),
        (sample_pytree, jnp.array([1, 2, 3]), 0, "", pytest.raises(ValueError)),
        (
            {"a": jnp.array([1, 2, 3]), "b": {"c": jnp.array([[1], [2]])}},
            jnp.array([1, 2]),
            0,
            "",
            pytest.raises(ValueError),
        ),
    ],
)
def test_check_array_shape_match_tree(tree, array, axis, err_message, expectation):
    with expectation:
        validation.check_array_shape_match_tree(tree, array, axis, err_message)


@pytest.mark.parametrize(
    "tree_1, tree_2, axis_1, axis_2, err_message, expectation",
    [
        (sample_pytree, sample_pytree, 0, 0, "", does_not_raise()),
        (
            {"a": jnp.array([1, 2, 3])},
            {"b": jnp.array([[1], [2]])},
            0,
            1,
            "Axis consistency mismatch.",
            pytest.raises(ValueError),
        ),
    ],
)
def test_check_tree_axis_consistency(
    tree_1, tree_2, axis_1, axis_2, err_message, expectation
):
    with expectation:
        validation.check_tree_axis_consistency(
            tree_1, tree_2, axis_1, axis_2, err_message
        )


@pytest.mark.parametrize(
    "tree_1, tree_2, err_message, expectation",
    [
        (sample_pytree, sample_pytree, "", does_not_raise()),
        (
            sample_pytree,
            {"a": [1, 2, 3], "b": {"c": [1, 2]}},
            "Tree structure mismatch.",
            pytest.raises(TypeError),
        ),
    ],
)
def test_check_tree_structure(tree_1, tree_2, err_message, expectation):
    with expectation:
        validation.check_tree_structure(tree_1, tree_2, err_message)



from ._basis import AdditiveBasis, MultiplicativeBasis
from ._transformer_basis import TransformerBasis
from .basis import (
    BSplineConv,
    BSplineEval,
    CyclicBSplineConv,
    CyclicBSplineEval,
    HistoryConv,
    IdentityEval,
    MSplineConv,
    MSplineEval,
    OrthExponentialConv,
    OrthExponentialEval,
    RaisedCosineLinearConv,
    RaisedCosineLinearEval,
    RaisedCosineLogConv,
    RaisedCosineLogEval,
)



import warnings
from contextlib import nullcontext as does_not_raise

import jax
import jax.numpy as jnp
import numpy as np
import pytest
import scipy.stats as sts
import statsmodels.api as sm
from numba import njit

import nemos as nmo


@pytest.fixture()
def poisson_observations():
    return nmo.observation_models.PoissonObservations


@pytest.fixture()
def gamma_observations():
    return nmo.observation_models.GammaObservations


class TestPoissonObservations:
    @pytest.mark.parametrize("link_function", [jnp.exp, jax.nn.softplus, 1])
    def test_initialization_link_is_callable(self, link_function, poisson_observations):
        """Check that the observation model initializes when a callable is passed."""
        raise_exception = not callable(link_function)
        if raise_exception:
            with pytest.raises(
                TypeError,
                match="The `inverse_link_function` function must be a Callable",
            ):
                poisson_observations(link_function)
        else:
            poisson_observations(link_function)

    @pytest.mark.parametrize(
        "link_function", [jnp.exp, np.exp, lambda x: x, sm.families.links.Log()]
    )
    def test_initialization_link_is_jax(self, link_function, poisson_observations):
        """Check that the observation model initializes when a callable is passed."""
        raise_exception = isinstance(link_function, np.ufunc) | isinstance(
            link_function, sm.families.links.Link
        )
        if raise_exception:
            with pytest.raises(
                TypeError,
                match="The `inverse_link_function` must return a jax.numpy.ndarray",
            ):
                poisson_observations(link_function)
        else:
            poisson_observations(link_function)

    @pytest.mark.parametrize("link_function", [jnp.exp, jax.nn.softplus, 1])
    def test_initialization_link_is_callable_set_params(
        self, link_function, poisson_observations
    ):
        """Check that the observation model initializes when a callable is passed."""
        observation_model = poisson_observations()
        raise_exception = not callable(link_function)
        if raise_exception:
            with pytest.raises(
                TypeError,
                match="The `inverse_link_function` function must be a Callable",
            ):
                observation_model.set_params(inverse_link_function=link_function)
        else:
            observation_model.set_params(inverse_link_function=link_function)

    @pytest.mark.parametrize(
        "link_function", [jnp.exp, np.exp, lambda x: x, sm.families.links.Log()]
    )
    def test_initialization_link_is_jax_set_params(
        self, link_function, poisson_observations
    ):
        """Check that the observation model initializes when a callable is passed."""
        raise_exception = isinstance(link_function, np.ufunc) | isinstance(
            link_function, sm.families.links.Link
        )
        observation_model = poisson_observations()
        if raise_exception:
            with pytest.raises(
                TypeError,
                match="The `inverse_link_function` must return a jax.numpy.ndarray!",
            ):
                observation_model.set_params(inverse_link_function=link_function)
        else:
            observation_model.set_params(inverse_link_function=link_function)

    @pytest.mark.parametrize(
        "link_function",
        [
            jnp.exp,
            lambda x: jnp.exp(x) if isinstance(x, jnp.ndarray) else "not a number",
        ],
    )
    def test_initialization_link_returns_scalar(
        self, link_function, poisson_observations
    ):
        """Check that the observation model initializes when a callable is passed."""
        raise_exception = not isinstance(link_function(1.0), (jnp.ndarray, float))
        observation_model = poisson_observations()
        if raise_exception:
            with pytest.raises(
                TypeError,
                match="The `inverse_link_function` must handle scalar inputs correctly",
            ):
                observation_model.set_params(inverse_link_function=link_function)
        else:
            observation_model.set_params(inverse_link_function=link_function)

    def test_get_params(self, poisson_observations):
        """Test get_params() returns expected values."""
        observation_model = poisson_observations()

        assert observation_model.get_params() == {
            "inverse_link_function": observation_model.inverse_link_function
        }

    def test_deviance_against_statsmodels(self, poissonGLM_model_instantiation):
        """
        Compare fitted parameters to statsmodels.
        Assesses if the model estimates are close to statsmodels' results.
        """
        _, y, model, _, firing_rate = poissonGLM_model_instantiation
        dev = sm.families.Poisson().deviance(y, firing_rate)
        dev_model = model.observation_model.deviance(y, firing_rate).sum()
        if not np.allclose(dev, dev_model):
            raise ValueError("Deviance doesn't match statsmodels!")

    def test_loglikelihood_against_scipy(self, poissonGLM_model_instantiation):
        """
        Compare log-likelihood to scipy.
        Assesses if the model estimates are close to statsmodels' results.
        """
        _, y, model, _, firing_rate = poissonGLM_model_instantiation
        ll_model = model.observation_model.log_likelihood(y, firing_rate)
        ll_scipy = sts.poisson(firing_rate).logpmf(y).mean()
        if not np.allclose(ll_model, ll_scipy):
            raise ValueError("Log-likelihood doesn't match scipy!")

    @pytest.mark.parametrize("score_type", ["pseudo-r2-Cohen", "pseudo-r2-McFadden"])
    def test_pseudo_r2_range(self, score_type, poissonGLM_model_instantiation):
        """
        Compute the pseudo-r2 and check that is < 1.
        """
        _, y, model, _, firing_rate = poissonGLM_model_instantiation
        pseudo_r2 = model.observation_model.pseudo_r2(
            y, firing_rate, score_type=score_type
        )
        if (pseudo_r2 > 1) or (pseudo_r2 < 0):
            raise ValueError(f"pseudo-r2 of {pseudo_r2} outside the [0,1] range!")

    @pytest.mark.parametrize("score_type", ["pseudo-r2-Cohen", "pseudo-r2-McFadden"])
    def test_pseudo_r2_mean(self, score_type, poissonGLM_model_instantiation):
        """
        Check that the pseudo-r2 of the null model is 0.
        """
        _, y, model, _, _ = poissonGLM_model_instantiation
        pseudo_r2 = model.observation_model.pseudo_r2(
            y, y.mean(), score_type=score_type
        )
        if not np.allclose(pseudo_r2, 0, atol=10**-7, rtol=0.0):
            raise ValueError(
                f"pseudo-r2 of {pseudo_r2} for the null model. Should be equal to 0!"
            )

    def test_emission_probability(selfself, poissonGLM_model_instantiation):
        """
        Test the poisson emission probability.

        Check that the emission probability is set to jax.random.poisson.
        """
        _, _, model, _, _ = poissonGLM_model_instantiation
        key_array = jax.random.key(123)
        counts = model.observation_model.sample_generator(key_array, np.arange(1, 11))
        if not jnp.all(counts == jax.random.poisson(key_array, np.arange(1, 11))):
            raise ValueError(
                "The emission probability should output the results of a call to jax.random.poisson."
            )

    @pytest.mark.parametrize(
        "score_type, expectation",
        [
            ("pseudo-r2-McFadden", does_not_raise()),
            (
                "not-implemented",
                pytest.raises(
                    NotImplementedError, match="Score not-implemented not implemented"
                ),
            ),
        ],
    )
    def test_not_implemented_score(
        self, score_type, expectation, poissonGLM_model_instantiation
    ):
        _, y, model, _, firing_rate = poissonGLM_model_instantiation
        with expectation:
            model.observation_model.pseudo_r2(y, firing_rate, score_type)

    @pytest.mark.parametrize(
        "scale, expectation",
        [
            (1, does_not_raise()),
            (
                "invalid",
                pytest.raises(
                    ValueError, match="The `scale` parameter must be of numeric type"
                ),
            ),
        ],
    )
    def test_scale_setter(self, scale, expectation, poissonGLM_model_instantiation):
        _, _, model, _, firing_rate = poissonGLM_model_instantiation
        with expectation:
            model.observation_model.scale = scale

    def test_scale_getter(self, poissonGLM_model_instantiation):
        _, _, model, _, firing_rate = poissonGLM_model_instantiation
        assert model.observation_model.scale == 1

    def test_non_differentiable_inverse_link(self, poissonGLM_model_instantiation):
        _, _, model, _, _ = poissonGLM_model_instantiation

        # define a jax non-diff function
        non_diff = lambda y: jnp.asarray(njit(lambda x: x)(np.atleast_1d(y)))

        with pytest.raises(
            TypeError,
            match="The `inverse_link_function` function cannot be differentiated",
        ):
            model.observation_model.inverse_link_function = non_diff

    def test_pseudo_r2_vs_statsmodels(self, poissonGLM_model_instantiation):
        """
        Compare log-likelihood to scipy.
        Assesses if the model estimates are close to statsmodels' results.
        """
        X, y, model, _, firing_rate = poissonGLM_model_instantiation

        # statsmodels mcfadden
        mdl = sm.GLM(y, sm.add_constant(X), family=sm.families.Poisson()).fit()
        pr2_sms = mdl.pseudo_rsquared("mcf")

        # set params
        pr2_model = model.observation_model.pseudo_r2(
            y, mdl.mu, score_type="pseudo-r2-McFadden"
        )

        if not np.allclose(pr2_model, pr2_sms):
            raise ValueError("Log-likelihood doesn't match statsmodels!")

    def test_aggregation_score_neg_ll(self, poissonGLM_model_instantiation):
        X, y, model, _, firing_rate = poissonGLM_model_instantiation
        sm = model.observation_model._negative_log_likelihood(y, firing_rate, jnp.sum)
        mn = model.observation_model._negative_log_likelihood(y, firing_rate, jnp.mean)
        assert np.allclose(sm, mn * y.shape[0])

    def test_aggregation_score_ll(self, poissonGLM_model_instantiation):
        X, y, model, _, firing_rate = poissonGLM_model_instantiation
        sm = model.observation_model.log_likelihood(
            y, firing_rate, aggregate_sample_scores=jnp.sum
        )
        mn = model.observation_model.log_likelihood(
            y, firing_rate, aggregate_sample_scores=jnp.mean
        )
        assert np.allclose(sm, mn * y.shape[0])

    @pytest.mark.parametrize("score_type", ["pseudo-r2-McFadden", "pseudo-r2-Cohen"])
    def test_aggregation_score_pr2(self, score_type, poissonGLM_model_instantiation):
        X, y, model, _, firing_rate = poissonGLM_model_instantiation
        sm = model.observation_model.pseudo_r2(
            y, firing_rate, score_type=score_type, aggregate_sample_scores=jnp.sum
        )
        mn = model.observation_model.pseudo_r2(
            y, firing_rate, score_type=score_type, aggregate_sample_scores=jnp.mean
        )
        assert np.allclose(sm, mn)

    def test_aggregation_score_mcfadden(self, poissonGLM_model_instantiation):
        X, y, model, _, firing_rate = poissonGLM_model_instantiation
        sm = model.observation_model._pseudo_r2_mcfadden(
            y, firing_rate, aggregate_sample_scores=jnp.sum
        )
        mn = model.observation_model._pseudo_r2_mcfadden(
            y, firing_rate, aggregate_sample_scores=jnp.mean
        )
        assert np.allclose(sm, mn)

    def test_aggregation_score_choen(self, poissonGLM_model_instantiation):
        X, y, model, _, firing_rate = poissonGLM_model_instantiation
        sm = model.observation_model._pseudo_r2_cohen(
            y, firing_rate, aggregate_sample_scores=jnp.sum
        )
        mn = model.observation_model._pseudo_r2_cohen(
            y, firing_rate, aggregate_sample_scores=jnp.mean
        )
        assert np.allclose(sm, mn)

    @pytest.mark.parametrize(
        "link_func, link_func_name", [(jnp.exp, "exp"), (jax.nn.softplus, "softplus")]
    )
    def test_repr_out(self, link_func, link_func_name):
        obs = nmo.observation_models.PoissonObservations(
            inverse_link_function=link_func
        )
        assert (
            repr(obs) == f"PoissonObservations(inverse_link_function={link_func_name})"
        )


class TestGammaObservations:
    @pytest.mark.parametrize("link_function", [jnp.exp, lambda x: 1 / x, 1])
    def test_initialization_link_is_callable(self, link_function, gamma_observations):
        """Check that the observation model initializes when a callable is passed."""
        raise_exception = not callable(link_function)
        if raise_exception:
            with pytest.raises(
                TypeError,
                match="The `inverse_link_function` function must be a Callable",
            ):
                gamma_observations(link_function)
        else:
            gamma_observations(link_function)

    @pytest.mark.parametrize(
        "link_function", [jnp.exp, np.exp, lambda x: 1 / x, sm.families.links.Log()]
    )
    def test_initialization_link_is_jax(self, link_function, gamma_observations):
        """Check that the observation model initializes when a callable is passed."""
        raise_exception = isinstance(link_function, np.ufunc) | isinstance(
            link_function, sm.families.links.Link
        )
        if raise_exception:
            with pytest.raises(
                TypeError,
                match="The `inverse_link_function` must return a jax.numpy.ndarray",
            ):
                gamma_observations(link_function)
        else:
            gamma_observations(link_function)

    @pytest.mark.parametrize("link_function", [jnp.exp, jax.nn.softplus, 1])
    def test_initialization_link_is_callable_set_params(
        self, link_function, gamma_observations
    ):
        """Check that the observation model initializes when a callable is passed."""
        observation_model = gamma_observations()
        raise_exception = not callable(link_function)
        if raise_exception:
            with pytest.raises(
                TypeError,
                match="The `inverse_link_function` function must be a Callable",
            ):
                observation_model.set_params(inverse_link_function=link_function)
        else:
            observation_model.set_params(inverse_link_function=link_function)

    @pytest.mark.parametrize(
        "link_function", [jnp.exp, np.exp, lambda x: x, sm.families.links.Log()]
    )
    def test_initialization_link_is_jax_set_params(
        self, link_function, gamma_observations
    ):
        """Check that the observation model initializes when a callable is passed."""
        raise_exception = isinstance(link_function, np.ufunc) | isinstance(
            link_function, sm.families.links.Link
        )
        observation_model = gamma_observations()
        if raise_exception:
            with pytest.raises(
                TypeError,
                match="The `inverse_link_function` must return a jax.numpy.ndarray!",
            ):
                observation_model.set_params(inverse_link_function=link_function)
        else:
            observation_model.set_params(inverse_link_function=link_function)

    @pytest.mark.parametrize(
        "link_function",
        [
            jnp.exp,
            lambda x: jnp.exp(x) if isinstance(x, jnp.ndarray) else "not a number",
        ],
    )
    def test_initialization_link_returns_scalar(
        self, link_function, gamma_observations
    ):
        """Check that the observation model initializes when a callable is passed."""
        raise_exception = not isinstance(link_function(1.0), (jnp.ndarray, float))
        observation_model = gamma_observations()
        if raise_exception:
            with pytest.raises(
                TypeError,
                match="The `inverse_link_function` must handle scalar inputs correctly",
            ):
                observation_model.set_params(inverse_link_function=link_function)
        else:
            observation_model.set_params(inverse_link_function=link_function)

    def test_get_params(self, gamma_observations):
        """Test get_params() returns expected values."""
        observation_model = gamma_observations()

        assert observation_model.get_params() == {
            "inverse_link_function": observation_model.inverse_link_function
        }

    def test_deviance_against_statsmodels(self, gammaGLM_model_instantiation):
        """
        Compare fitted parameters to statsmodels.
        Assesses if the model estimates are close to statsmodels' results.
        """
        _, y, model, _, firing_rate = gammaGLM_model_instantiation
        dev = sm.families.Gamma().deviance(y, firing_rate)
        dev_model = model.observation_model.deviance(y, firing_rate).sum()
        if not np.allclose(dev, dev_model):
            raise ValueError("Deviance doesn't match statsmodels!")

    def test_loglikelihood_against_statsmodels(self, gammaGLM_model_instantiation):
        """
        Compare log-likelihood to scipy.
        Assesses if the model estimates are close to statsmodels' results.
        """
        _, y, model, _, firing_rate = gammaGLM_model_instantiation
        ll_model = model.observation_model.log_likelihood(y, firing_rate)
        ll_sms = sm.families.Gamma().loglike(y, firing_rate) / y.shape[0]
        if not np.allclose(ll_model, ll_sms):
            raise ValueError("Log-likelihood doesn't match statsmodels!")

    @pytest.mark.parametrize("score_type", ["pseudo-r2-Cohen", "pseudo-r2-McFadden"])
    def test_pseudo_r2_range(self, score_type, gammaGLM_model_instantiation):
        """
        Compute the pseudo-r2 and check that is < 1.
        """
        X, y, model, true_params, firing_rate = gammaGLM_model_instantiation
        model.coef_ = true_params[0]
        model.intercept_ = true_params[1]
        model.scale_ = 0.5

        rate = model.predict(X)
        ysim, _ = model.simulate(jax.random.PRNGKey(123), X)
        pseudo_r2 = nmo.observation_models.GammaObservations(
            inverse_link_function=lambda x: 1 / x
        ).pseudo_r2(ysim, rate, score_type=score_type)
        if (pseudo_r2 > 1) or (pseudo_r2 < 0):
            raise ValueError(f"pseudo-r2 of {pseudo_r2} outside the [0,1] range!")

    @pytest.mark.parametrize("score_type", ["pseudo-r2-Cohen", "pseudo-r2-McFadden"])
    def test_pseudo_r2_mean(self, score_type, gammaGLM_model_instantiation):
        """
        Check that the pseudo-r2 of the null model is 0.
        """
        _, y, model, _, _ = gammaGLM_model_instantiation
        pseudo_r2 = model.observation_model.pseudo_r2(
            y, y.mean(), score_type=score_type
        )
        if not np.allclose(pseudo_r2, 0, atol=10**-7, rtol=0.0):
            raise ValueError(
                f"pseudo-r2 of {pseudo_r2} for the null model. Should be equal to 0!"
            )

    def test_emission_probability(self, gammaGLM_model_instantiation):
        """
        Test the gamma emission probability.

        Check that the emission probability is set to jax.random.gamma.
        """
        _, _, model, _, _ = gammaGLM_model_instantiation
        key_array = jax.random.key(123)
        counts = model.observation_model.sample_generator(key_array, np.arange(1, 11))
        if not jnp.all(counts == jax.random.gamma(key_array, np.arange(1, 11))):
            raise ValueError(
                "The emission probability should output the results of a call to jax.random.gamma."
            )

    @pytest.mark.parametrize(
        "score_type, expectation",
        [
            ("pseudo-r2-McFadden", does_not_raise()),
            (
                "not-implemented",
                pytest.raises(
                    NotImplementedError, match="Score not-implemented not implemented"
                ),
            ),
        ],
    )
    def test_not_implemented_score(
        self, score_type, expectation, gammaGLM_model_instantiation
    ):
        _, y, model, _, firing_rate = gammaGLM_model_instantiation
        with expectation:
            model.observation_model.pseudo_r2(y, firing_rate, score_type)

    @pytest.mark.parametrize(
        "scale, expectation",
        [
            (1, does_not_raise()),
            (
                "invalid",
                pytest.raises(
                    ValueError, match="The `scale` parameter must be of numeric type"
                ),
            ),
        ],
    )
    def test_scale_setter(self, scale, expectation, gammaGLM_model_instantiation):
        _, _, model, _, firing_rate = gammaGLM_model_instantiation
        with expectation:
            model.observation_model.scale = scale

    def test_scale_getter(self, gammaGLM_model_instantiation):
        _, _, model, _, firing_rate = gammaGLM_model_instantiation
        assert model.observation_model.scale == 1

    def test_non_differentiable_inverse_link(self, gammaGLM_model_instantiation):
        _, _, model, _, _ = gammaGLM_model_instantiation

        # define a jax non-diff function
        non_diff = lambda y: jnp.asarray(njit(lambda x: x)(np.atleast_1d(y)))

        with pytest.raises(
            TypeError,
            match="The `inverse_link_function` function cannot be differentiated",
        ):
            model.observation_model.inverse_link_function = non_diff

    def test_pseudo_r2_vs_statsmodels(self, gammaGLM_model_instantiation):
        """
        Compare log-likelihood to scipy.
        Assesses if the model estimates are close to statsmodels' results.
        """
        jax.config.update("jax_enable_x64", True)
        X, y, model, _, firing_rate = gammaGLM_model_instantiation

        # statsmodels mcfadden
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", message="The InversePower link function does"
            )
            mdl = sm.GLM(y, sm.add_constant(X), family=sm.families.Gamma()).fit()
        pr2_sms = mdl.pseudo_rsquared("mcf")

        # set params
        pr2_model = model.observation_model.pseudo_r2(
            y, mdl.mu, score_type="pseudo-r2-McFadden", scale=mdl.scale
        )

        if not np.allclose(pr2_model, pr2_sms):
            raise ValueError("Log-likelihood doesn't match statsmodels!")

    def test_aggregation_score_neg_ll(self, gammaGLM_model_instantiation):
        X, y, model, _, firing_rate = gammaGLM_model_instantiation
        sm = model.observation_model._negative_log_likelihood(y, firing_rate, jnp.sum)
        mn = model.observation_model._negative_log_likelihood(y, firing_rate, jnp.mean)
        assert np.allclose(sm, mn * y.shape[0])

    def test_aggregation_score_ll(self, gammaGLM_model_instantiation):
        X, y, model, _, firing_rate = gammaGLM_model_instantiation
        sm = model.observation_model.log_likelihood(
            y, firing_rate, aggregate_sample_scores=jnp.sum
        )
        mn = model.observation_model.log_likelihood(
            y, firing_rate, aggregate_sample_scores=jnp.mean
        )
        assert np.allclose(sm, mn * y.shape[0])

    @pytest.mark.parametrize("score_type", ["pseudo-r2-McFadden", "pseudo-r2-Cohen"])
    def test_aggregation_score_pr2(self, score_type, gammaGLM_model_instantiation):
        X, y, model, _, firing_rate = gammaGLM_model_instantiation
        sm = model.observation_model.pseudo_r2(
            y, firing_rate, score_type=score_type, aggregate_sample_scores=jnp.sum
        )
        mn = model.observation_model.pseudo_r2(
            y, firing_rate, score_type=score_type, aggregate_sample_scores=jnp.mean
        )
        assert np.allclose(sm, mn)

    def test_aggregation_score_mcfadden(self, gammaGLM_model_instantiation):
        X, y, model, _, firing_rate = gammaGLM_model_instantiation
        sm = model.observation_model._pseudo_r2_mcfadden(
            y, firing_rate, aggregate_sample_scores=jnp.sum
        )
        mn = model.observation_model._pseudo_r2_mcfadden(
            y, firing_rate, aggregate_sample_scores=jnp.mean
        )
        assert np.allclose(sm, mn)

    def test_aggregation_score_choen(self, gammaGLM_model_instantiation):
        X, y, model, _, firing_rate = gammaGLM_model_instantiation
        sm = model.observation_model._pseudo_r2_cohen(
            y, firing_rate, aggregate_sample_scores=jnp.sum
        )
        mn = model.observation_model._pseudo_r2_cohen(
            y, firing_rate, aggregate_sample_scores=jnp.mean
        )
        assert np.allclose(sm, mn)

    @pytest.mark.parametrize(
        "link_func, link_func_name", [(jnp.exp, "exp"), (jax.nn.softplus, "softplus")]
    )
    def test_repr_out(self, link_func, link_func_name):
        obs = nmo.observation_models.GammaObservations(inverse_link_function=link_func)
        assert repr(obs) == f"GammaObservations(inverse_link_function={link_func_name})"



# required to get ArrayLike to render correctly
from __future__ import annotations

import abc
import copy
from typing import Literal, Optional, Tuple

import numpy as np
from numpy.typing import ArrayLike, NDArray
from pynapple import Tsd, TsdFrame, TsdTensor
from scipy.interpolate import splev

from ..type_casting import support_pynapple
from ..typing import FeatureMatrix
from ._basis import Basis, check_transform_input, min_max_rescale_samples
from ._basis_mixin import AtomicBasisMixin


class SplineBasis(Basis, AtomicBasisMixin, abc.ABC):
    """
    SplineBasis class inherits from the Basis class and represents spline basis functions.

    Parameters
    ----------
    n_basis_funcs :
        Number of basis functions.
    mode :
        The mode of operation. 'eval' for evaluation at sample points,
        'conv' for convolutional operation.
    order : optional
        Spline order.
    label :
        The label of the basis, intended to be descriptive of the task variable being processed.
        For example: velocity, position, spike_counts.

    Attributes
    ----------
    order : int
        Spline order.
    """

    def __init__(
        self,
        n_basis_funcs: int,
        order: int = 2,
        label: Optional[str] = None,
        mode: Literal["conv", "eval"] = "eval",
    ) -> None:
        self.order = order
        AtomicBasisMixin.__init__(self, n_basis_funcs=n_basis_funcs)
        super().__init__(
            label=label,
            mode=mode,
        )

        self._n_input_dimensionality = 1

    @property
    def order(self):
        """
        Spline order.

        Spline order, i.e. the polynomial degree of the spline plus one.
        """
        return self._order

    @order.setter
    def order(self, value):
        """Setter for the order parameter."""
        if value != int(value):
            raise ValueError(
                f"Spline order must be an integer! Order {value} provided."
            )
        value = int(value)
        if value < 1:
            raise ValueError(f"Spline order must be positive! Order {value} provided.")

        # Set to None only the first time the setter is called.
        orig_order = copy.deepcopy(getattr(self, "_order", None))

        # Set the order
        self._order = value

        # If the order was already initialized, re-check basis
        if orig_order is not None:
            try:
                self._check_n_basis_min()
            except ValueError as e:
                self._order = orig_order
                raise e

    def _generate_knots(
        self,
        is_cyclic: bool = False,
    ) -> NDArray:
        """
        Generate knots locations for spline basis functions.

        Parameters
        ----------
        is_cyclic : optional
            Whether the spline is cyclic.

        Returns
        -------
            The knot locations for the spline basis functions.

        Raises
        ------
        AssertionError
            If the percentiles or order values are not within the valid range.
        """
        # Determine number of interior knots.
        num_interior_knots = self.n_basis_funcs - self.order
        if is_cyclic:
            num_interior_knots += self.order - 1

        # Spline basis have support on the semi-open [a, b)  interval, we add a small epsilon
        # to mx so that the so that basis_element(max(samples)) != 0
        knot_locs = np.concatenate(
            (
                np.zeros(self.order - 1),
                np.linspace(0, (1 + np.finfo(float).eps), num_interior_knots + 2),
                np.full(self.order - 1, 1 + np.finfo(float).eps),
            )
        )
        return knot_locs

    def _check_n_basis_min(self) -> None:
        """Check that the user required enough basis elements.

        Check that the spline-basis has at least as many basis as the order.

        Raises
        ------
        ValueError
            If an insufficient number of basis element is requested for the basis type
        """
        if self.n_basis_funcs < self.order:
            raise ValueError(
                f"{self.__class__.__name__} `order` parameter cannot be larger "
                "than `n_basis_funcs` parameter."
            )


class MSplineBasis(SplineBasis, abc.ABC):
    r"""
    M-spline basis functions for modeling and data transformation.

    M-splines [1]_ are a type of spline basis function used for smooth curve fitting
    and data representation. They are positive and integrate to one, making them
    suitable for probabilistic models and density estimation. The order of an
    M-spline defines its smoothness, with higher orders resulting in smoother
    splines.

    This class provides functionality to create M-spline basis functions, allowing
    for flexible and smooth modeling of data. It inherits from the ``SplineBasis``
    abstract class, providing specific implementations for M-splines.

    Parameters
    ----------
    n_basis_funcs :
        The number of basis functions to generate. More basis functions allow for
        more flexible data modeling but can lead to overfitting.
    mode :
        The mode of operation. 'eval' for evaluation at sample points,
        'conv' for convolutional operation.
    order :
        The order of the splines used in basis functions. Must be between [1,
        n_basis_funcs]. Default is 2. Higher order splines have more continuous
        derivatives at each interior knot, resulting in smoother basis functions.
    label :
        The label of the basis, intended to be descriptive of the task variable being processed.
        For example: velocity, position, spike_counts.

    References
    ----------
    .. [1] Ramsay, J. O. (1988). Monotone regression splines in action. Statistical science,
        3(4), 425-441.

    Notes
    -----
    ``MSplines`` must integrate to 1 over their domain (the area under the curve is 1). Therefore, if the domain
    (x-axis) of an MSpline basis is expanded by a factor of :math:`\alpha`, the values on the co-domain (y-axis) values
    will shrink by a factor of :math:`1/\alpha`.
    For example, over the standard bounds of (0, 1), the maximum value of the MSpline is 18.
    If we set the bounds to (0, 2), the maximum value will be 9, i.e., 18 / 2.

    Examples
    --------
    >>> from numpy import linspace
    >>> from nemos.basis import MSplineEval
    >>> n_basis_funcs = 5
    >>> order = 3
    >>> mspline_basis = MSplineEval(n_basis_funcs, order=order)
    >>> sample_points = linspace(0, 1, 100)
    >>> basis_functions = mspline_basis.compute_features(sample_points)
    """

    def __init__(
        self,
        n_basis_funcs: int,
        mode: Literal["eval", "conv"] = "eval",
        order: int = 2,
        label: Optional[str] = "MSplineEval",
    ) -> None:
        super().__init__(
            n_basis_funcs,
            mode=mode,
            order=order,
            label=label,
        )

    @support_pynapple(conv_type="numpy")
    @check_transform_input
    def _evaluate(
        self, sample_pts: ArrayLike | Tsd | TsdFrame | TsdTensor
    ) -> FeatureMatrix:
        """
        Evaluate the M-spline basis functions at given sample points.

        Parameters
        ----------
        sample_pts :
            The sample points at which the M-spline is evaluated.
            `sample_pts` is a n-dimensional (n >= 1) array with first axis being the samples, i.e.
            `sample_pts.shape[0] == n_samples`.

        Returns
        -------
        :
            An array where each column corresponds to one M-spline basis function
            evaluated at the input sample points. The shape of the array is
            (len(sample_pts), n_basis_funcs).

        Notes
        -----
        The implementation uses a recursive definition of M-splines. Boundary
        conditions are handled such that the basis functions are positive and
        integrate to one over the domain defined by the sample points.
        """
        sample_pts, scaling = min_max_rescale_samples(
            sample_pts, getattr(self, "bounds", None)
        )
        # add knots if not passed
        knot_locs = self._generate_knots(is_cyclic=False)

        # get the original shape
        shape = sample_pts.shape
        X = np.stack(
            [
                mspline(
                    sample_pts.reshape(
                        -1,
                    ),
                    self.order,
                    i,
                    knot_locs,
                )
                for i in range(self.n_basis_funcs)
            ],
            axis=1,
        )
        X = X.reshape(*shape, X.shape[1])
        # re-normalize so that it integrates to 1 over the range.
        X /= scaling[..., None]

        return X

    def evaluate_on_grid(self, n_samples: int) -> Tuple[NDArray, NDArray]:
        """
        Evaluate the M-spline basis functions on a uniformly spaced grid.

        This method creates a uniformly spaced grid of sample points within the domain
        [0, 1] and evaluates all the M-spline basis functions at these points. It is
        particularly useful for visualizing the shape and distribution of the basis
        functions across their domain.

        Parameters
        ----------
        n_samples :
            The number of points in the uniformly spaced grid. A higher number of
            samples will result in a more detailed visualization of the basis functions.

        Returns
        -------
        X : NDArray
            A 1D array of uniformly spaced sample points within the domain [0, 1].
            Shape: ``(n_samples,)``.
        Y : NDArray
            A 2D array where each row corresponds to the evaluated M-spline basis
            function values at the points in X. Shape: ``(n_samples, n_basis_funcs)``.
        """
        return super().evaluate_on_grid(n_samples)


class BSplineBasis(SplineBasis, abc.ABC):
    """
    B-spline 1-dimensional basis functions.

    Implementation of the one-dimensional BSpline basis [1]_.

    Parameters
    ----------
    n_basis_funcs :
        Number of basis functions.
    mode :
        The mode of operation. ``'eval'`` for evaluation at sample points,
        'conv' for convolutional operation.
    order :
        Order of the splines used in basis functions. Must lie within ``[1, n_basis_funcs]``.
        The B-splines have (order-2) continuous derivatives at each interior knot.
        The higher this number, the smoother the basis representation will be.
    label :
        The label of the basis, intended to be descriptive of the task variable being processed.
        For example: velocity, position, spike_counts.

    Attributes
    ----------
    order :
        Spline order.


    References
    ----------
    .. [1] Prautzsch, H., Boehm, W., Paluszny, M. (2002). B-spline representation. In: Bézier and B-Spline Techniques.
        Mathematics and Visualization. Springer, Berlin, Heidelberg. https://doi.org/10.1007/978-3-662-04919-8_5
    """

    def __init__(
        self,
        n_basis_funcs: int,
        mode="eval",
        order: int = 4,
        label: Optional[str] = "BSplineBasis",
    ):
        super().__init__(
            n_basis_funcs,
            mode=mode,
            order=order,
            label=label,
        )

    @support_pynapple(conv_type="numpy")
    @check_transform_input
    def _evaluate(
        self, sample_pts: ArrayLike | Tsd | TsdFrame | TsdTensor
    ) -> FeatureMatrix:
        """
        Evaluate the B-spline basis functions with given sample points.

        Parameters
        ----------
        sample_pts :
            The sample points at which the B-spline is evaluated, shape (n_samples,).
            `sample_pts` is a n-dimensional (n >= 1) array with first axis being the samples, i.e.
            `sample_pts.shape[0] == n_samples`.

        Returns
        -------
        basis_funcs :
            The basis function evaluated at the samples, shape (n_samples, n_basis_funcs).

        Raises
        ------
        AssertionError
            If the sample points are not within the B-spline knots.

        Notes
        -----
        The evaluation is performed by looping over each element and using ``splev``
        from SciPy to compute the basis values.
        """
        sample_pts, _ = min_max_rescale_samples(
            sample_pts, getattr(self, "bounds", None)
        )
        # add knots
        knot_locs = self._generate_knots(is_cyclic=False)

        # reshape to flat and store original shape
        shape = sample_pts.shape
        sample_pts = sample_pts.reshape(
            -1,
        )

        basis_eval = bspline(
            sample_pts, knot_locs, order=self.order, der=0, outer_ok=False
        )
        basis_eval = basis_eval.reshape(*shape, basis_eval.shape[1])
        return basis_eval

    def evaluate_on_grid(self, n_samples: int) -> Tuple[NDArray, NDArray]:
        """Evaluate the B-spline basis set on a grid of equi-spaced sample points.

        Parameters
        ----------
        n_samples :
            The number of points in the uniformly spaced grid. A higher number of
            samples will result in a more detailed visualization of the basis functions.

        Returns
        -------
        X :
            Array of shape ``(n_samples,)`` containing the equi-spaced sample
            points where we've evaluated the basis.
        basis_funcs :
            Raised cosine basis functions, shape ``(n_samples, n_basis_funcs)``

        Notes
        -----
        The evaluation is performed by looping over each element and using ``splev`` from
        SciPy to compute the basis values.
        """
        return super().evaluate_on_grid(n_samples)


class CyclicBSplineBasis(SplineBasis, abc.ABC):
    """
    B-spline 1-dimensional basis functions for cyclic splines.

    Parameters
    ----------
    n_basis_funcs :
        Number of basis functions.
    mode :
        The mode of operation. 'eval' for evaluation at sample points,
        'conv' for convolutional operation.
    order :
        Order of the splines used in basis functions. Order must lie within [2, n_basis_funcs].
        The B-splines have (order-2) continuous derivatives at each interior knot.
        The higher this number, the smoother the basis representation will be.
    label :
        The label of the basis, intended to be descriptive of the task variable being processed.
        For example: velocity, position, spike_counts.

    Attributes
    ----------
    n_basis_funcs :
        Number of basis functions, int.
    order :
        Order of the splines used in basis functions, int.
    """

    def __init__(
        self,
        n_basis_funcs: int,
        mode="eval",
        order: int = 4,
        label: Optional[str] = "CyclicBSplineBasis",
    ):
        super().__init__(
            n_basis_funcs,
            mode=mode,
            order=order,
            label=label,
        )
        if self.order < 2:
            raise ValueError(
                f"Order >= 2 required for cyclic B-spline, "
                f"order {self.order} specified instead!"
            )

    @support_pynapple(conv_type="numpy")
    @check_transform_input
    def _evaluate(
        self,
        sample_pts: ArrayLike | Tsd | TsdFrame | TsdTensor,
    ) -> FeatureMatrix:
        """Evaluate the Cyclic B-spline basis functions with given sample points.

        Parameters
        ----------
        sample_pts :
            The sample points at which the cyclic B-spline is evaluated. Samples must be stored in a
            multi-dimensional array with first axis being the samples, i.e. `sample_pts.shape[0] == n_samples`.

        Returns
        -------
        basis_funcs :
            The basis function evaluated at the samples, shape (n_samples, n_basis_funcs)

        Notes
        -----
        The evaluation is performed by looping over each element and using ``splev`` from
        SciPy to compute the basis values.

        """
        sample_pts, _ = min_max_rescale_samples(
            sample_pts, getattr(self, "bounds", None)
        )
        knot_locs = self._generate_knots(is_cyclic=True)

        # for cyclic, do not repeat knots
        knot_locs = np.unique(knot_locs)

        nk = knot_locs.shape[0]

        # make sure knots are sorted
        knot_locs.sort()

        # extend knots
        xc = knot_locs[nk - self.order]
        knots = np.hstack(
            (
                knot_locs[0] - knot_locs[-1] + knot_locs[nk - self.order : nk - 1],
                knot_locs,
            )
        )

        # reshape to flat and store original shape
        shape = sample_pts.shape
        sample_pts = sample_pts.reshape(
            -1,
        )

        ind = sample_pts > xc

        basis_eval = bspline(sample_pts, knots, order=self.order, der=0, outer_ok=True)
        sample_pts[ind] = sample_pts[ind] - knots.max() + knot_locs[0]

        if np.sum(ind):
            basis_eval[ind] = basis_eval[ind] + bspline(
                sample_pts[ind], knots, order=self.order, outer_ok=True, der=0
            )
        # restore points
        sample_pts[ind] = sample_pts[ind] + knots.max() - knot_locs[0]

        basis_eval = basis_eval.reshape(*shape, basis_eval.shape[1])
        return basis_eval

    def evaluate_on_grid(self, n_samples: int) -> Tuple[NDArray, NDArray]:
        """Evaluate the Cyclic B-spline basis set on a grid of equi-spaced sample points.

        Parameters
        ----------
        n_samples :
            The number of points in the uniformly spaced grid. A higher number of
            samples will result in a more detailed visualization of the basis functions.

        Returns
        -------
        X :
            Array of shape ``(n_samples,)`` containing the equi-spaced sample
            points where we've evaluated the basis.
        basis_funcs :
            Raised cosine basis functions, shape ``(n_samples, n_basis_funcs)``

        Notes
        -----
        The evaluation is performed by looping over each element and using ``splev`` from
        SciPy to compute the basis values.
        """
        return super().evaluate_on_grid(n_samples)


def mspline(x: NDArray, k: int, i: int, T: NDArray) -> NDArray:
    """Compute M-spline basis function.

    Parameters
    ----------
    x
        Spacing for basis functions, shape (n_sample_points, ).
    k
        Order of the spline basis.
    i
        Number of the spline basis.
    T
        knot locations. should lie in interval [0, 1], shape (k + n_basis_funcs,).

    Returns
    -------
    spline
        M-spline basis function, shape (n_sample_points, ).

    Examples
    --------
    >>> import numpy as np
    >>> from numpy import linspace
    >>> from nemos.basis._spline_basis import mspline
    >>> sample_points = linspace(0, 1, 100)
    >>> mspline_eval = mspline(x=sample_points, k=3, i=2, T=np.random.rand(7)) # define a cubic M-spline
    >>> mspline_eval.shape
    (100,)
    """
    # Boundary conditions.
    if (T[i + k] - T[i]) < 1e-6:
        return np.zeros_like(x)

    # Special base case of first-order spline basis.
    elif k == 1:
        v = np.zeros_like(x)
        v[(x >= T[i]) & (x < T[i + 1])] = 1 / (T[i + 1] - T[i])
        return v

    # General case, defined recursively
    else:
        return (
            k
            * (
                (x - T[i]) * mspline(x, k - 1, i, T)
                + (T[i + k] - x) * mspline(x, k - 1, i + 1, T)
            )
            / ((k - 1) * (T[i + k] - T[i]))
        )


def bspline(
    sample_pts: NDArray,
    knots: NDArray,
    order: int = 4,
    der: int = 0,
    outer_ok: bool = False,
) -> NDArray:
    """
    Calculate and return the evaluation of B-spline basis.

    This function evaluates B-spline basis for given sample points. It checks for
    out of range points and optionally handles them. It also handles the NaNs if present.

    Parameters
    ----------
    sample_pts :
        An array containing sample points for which B-spline basis needs to be evaluated,
        shape (n_samples,)
    knots :
        An array containing knots for the B-spline basis. The knots are sorted in ascending order.
    order :
        The order of the B-spline basis.
    der :
        The derivative of the B-spline basis to be evaluated.
    outer_ok :
        If True, allows for evaluation at points outside the range of knots.
        Default is False, in which case an assertion error is raised when
        points outside the knots range are encountered.

    Returns
    -------
    basis_eval :
        An array containing the evaluation of B-spline basis for the given sample points.
        Shape (n_samples, n_basis_funcs).

    Raises
    ------
    AssertionError
        If `outer_ok` is False and the sample points lie outside the B-spline knots range.

    Notes
    -----
    The function uses splev function from scipy.interpolate library for the basis evaluation.

    Examples
    --------
    >>> import numpy as np
    >>> from numpy import linspace
    >>> from nemos.basis._spline_basis import bspline
    >>> sample_points = linspace(0, 1, 100)
    >>> knots = np.array([0, 0, 0, 0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 1, 1, 1, 1])
    >>> bspline_eval = bspline(sample_points, knots) # define a cubic B-spline
    >>> bspline_eval.shape
    (100, 10)
    """
    knots.sort()
    nk = knots.shape[0]

    # check for out of range points (in cyclic b-spline need_outer must be set to False)
    need_outer = any(sample_pts < knots[order - 1]) or any(
        sample_pts > knots[nk - order]
    )
    assert (
        not need_outer
    ) | outer_ok, 'sample points must lie within the B-spline knots range unless "outer_ok==True".'

    # select knots that are within the knots range (this takes care of eventual NaNs)
    in_sample = (sample_pts >= knots[0]) & (sample_pts <= knots[-1])

    if need_outer:
        reps = order - 1
        knots = np.hstack((np.ones(reps) * knots[0], knots, np.ones(reps) * knots[-1]))
        nk = knots.shape[0]
    else:
        reps = 0

    # number of basis elements
    n_basis = nk - order

    # initialize the basis element container
    basis_eval = np.full((n_basis - 2 * reps, sample_pts.shape[0]), np.nan)

    # loop one element at the time and evaluate the basis using splev
    id_basis = np.eye(n_basis, nk, dtype=np.int8)
    for i in range(reps, len(knots) - order - reps):
        basis_eval[i - reps, in_sample] = splev(
            sample_pts[in_sample], (knots, id_basis[i], order - 1), der=der
        )

    return basis_eval.T



"""Utility functions for data pre-processing."""

import inspect
import os
import warnings
from typing import Any, Callable, List, Literal, Optional, Union

import jax
import jax.numpy as jnp
import numpy as np
from numpy.typing import NDArray

from .base_class import Base
from .tree_utils import pytree_map_and_reduce
from .type_casting import is_numpy_array_like, support_pynapple


def check_dimensionality(
    pytree: Any,
    expected_dim: int,
) -> bool:
    """
    Check the dimensionality of the arrays in a pytree.

    Check that all arrays in pytree have the expected dimensionality.

    Parameters
    ----------
    pytree :
        A pytree object.
    expected_dim :
        Number of expected dimension for the NDArrays.

    Returns
    -------
    True if all the arrays has the expected number of dimension, False otherwise.
    """
    return not pytree_map_and_reduce(lambda x: x.ndim != expected_dim, any, pytree)


def validate_axis(tree: Any, axis: int):
    """
    Validate the axis for each array in a given tree structure.

    This function checks if the specified axis exists in each array within the tree. It raises a ValueError
    if the specified axis is equal to or greater than the number of dimensions in any of the arrays.

    Parameters
    ----------
    tree :
        A tree containing arrays.
    axis :
        The axis that should be valid for each array in the tree. This means each array must have at least
        `axis + 1` dimensions.

    Raises
    ------
    ValueError
        If the specified axis is equal to or greater than the number of dimensions (`ndim`) of any array
        within the tree. This ensures that operations intended for a specific axis can be safely performed
        on every array in the tree.
    ValueError
        If the axis is negative or non-integer.
    """
    if not isinstance(axis, int) or axis < 0:
        raise ValueError("`axis` must be a non negative integer.")

    if pytree_map_and_reduce(lambda x: x.ndim <= axis, any, tree):
        raise ValueError(
            "'axis' must be smaller than the number of dimensions of any array in 'tree'."
        )


def check_non_empty(pytree: Any, pytree_name: str):
    """
    Check if any array in the pytree is empty.

    Parameters
    ----------
    pytree :
        A pytree object containing arrays.
    pytree_name :
        The name of the pytree variable for error message purposes.

    Raises
    ------
    ValueError
        If any array in the pytree is empty (i.e., has a zero dimension).
    """
    if pytree_map_and_reduce(lambda x: 0 in x.shape, any, pytree):
        raise ValueError(
            f"Empty array provided. At least one of dimension in {pytree_name} is empty."
        )


def check_trials_longer_than_time_window(
    time_series: Any, window_size: int, axis: int = 0
):
    """
    Check if the duration of each trial in the time series is at least as long as the window size.

    Parameters
    ----------
    time_series :
        A pytree of trial data.
    window_size :
        The size of the window to be used in convolution.
    axis :
        The axis in the arrays representing the time dimension.

    Raises
    ------
    ValueError
        If any trial in the time series is shorter than the window size.
    """
    # Check window size
    if pytree_map_and_reduce(lambda x: x.shape[axis] < window_size, any, time_series):
        raise ValueError(
            "Insufficient trial duration. The number of time points in each trial must "
            "be greater or equal to the window size."
        )


def _pad_dimension(
    array: jnp.ndarray,
    axis: int,
    pad_size: int,
    predictor_causality: Literal["causal", "acausal", "anti-causal"] = "causal",
    constant_values: float = jnp.nan,
) -> jnp.ndarray:
    """
    Add padding to the last dimension of an array based on the convolution type.

    This is a helper function used by `nan_pad_conv`, which is the function we expect the user will interact with.

    Parameters
    ----------
    array:
        The array to be padded.
    axis:
        The axis to be padded.
    pad_size:
        The number of NaNs to concatenate as padding.
    predictor_causality:
        Causality of this predictor, which determines where padded values are added.
    constant_values:
        The constant values for padding, default is jnp.nan.

    Returns
    -------
    :
        An array with padded last dimension.
    """
    padding_settings = {
        "causal": (pad_size, 0),
        "acausal": ((pad_size) // 2, pad_size - (pad_size) // 2),
        "anti-causal": (0, pad_size),
    }

    pad_width = (
        ((0, 0),) * axis
        + (padding_settings[predictor_causality],)
        + ((0, 0),) * (array.ndim - 1 - axis)
    )
    return jnp.pad(array, pad_width, constant_values=constant_values)


def nan_pad(
    conv_time_series: Any,
    pad_size: int,
    predictor_causality: Literal["causal", "acausal", "anti-causal"] = "causal",
    axis: int = 0,
) -> Any:
    """
    Add NaN padding to a convolved time series based on specified causality and axis.

    This function pads the convolved time series with NaNs along a specified axis. The amount
    and location of the padding are determined by the pad_size and predictor_causality parameters.

    Parameters
    ----------
    conv_time_series :
        The convolved time series to pad. This variable should be a pytree with arrays as leaves.
        The structure can be one of the following:
        1. A single array with ndim > axis.
        2. A pytree whose leaves are arrays.
    pad_size :
        The number of NaNs to concatenate as padding.
    predictor_causality : {'causal', 'acausal', 'anti-causal'}, default='causal'
        Causality of the predictor, which determines where padded values are added:
        - 'causal': Padding is added before the data.
        - 'acausal': Padding is evenly distributed before and after the data.
        - 'anti-causal': Padding is added after the data.
    axis : int, default=0
        The axis along which to add padding.

    Returns
    -------
    padded_conv_time_series : Any
        The convolved time series with NaN padding. The structure matches that of `conv_time_series`.

    Raises
    ------
    ValueError
        - If `pad_size` is not a positive integer.
        - If `predictor_causality` is not one of the expected values ('causal', 'acausal', 'anti-causal').
        - If `axis` is not a valid axis for any of the arrays in `conv_time_series`, specifically
          if `axis >= array.ndim` for any array.
        - If any array in `conv_time_series` does not have a floating-point data type.
    """
    if not isinstance(pad_size, int) or pad_size <= 0:
        raise ValueError(
            f"pad_size must be a positive integer! Pad size of {pad_size} provided instead!"
        )

    causality_choices = ["causal", "acausal", "anti-causal"]
    if predictor_causality not in causality_choices:
        raise ValueError(
            f"predictor_causality must be one of {causality_choices}. {predictor_causality} provided instead!"
        )
    if predictor_causality == "acausal" and (pad_size % 2 == 1):
        warnings.warn(
            "With acausal filter, pad_size should probably be even,"
            " so that we can place an equal number of NaNs on either side of input",
            UserWarning,
        )

    # convert to jax ndarray
    conv_time_series = jax.tree_util.tree_map(jnp.asarray, conv_time_series)

    # validate the axis
    validate_axis(conv_time_series, axis)

    if is_numpy_array_like(conv_time_series):
        if not np.issubdtype(conv_time_series.dtype, np.floating):
            raise ValueError("conv_time_series must have a float dtype!")
        return _pad_dimension(
            conv_time_series,
            axis,
            pad_size,
            predictor_causality,
            constant_values=jnp.nan,
        )

    else:
        if pytree_map_and_reduce(
            lambda trial: not np.issubdtype(trial.dtype, np.floating),
            any,
            conv_time_series,
        ):
            raise ValueError("All leaves of conv_time_series must have a float dtype!")
        return jax.tree_util.tree_map(
            lambda trial: _pad_dimension(
                trial, axis, pad_size, predictor_causality, constant_values=jnp.nan
            ),
            conv_time_series,
        )


def _compute_index_adjust(
    time_series: NDArray, causality: Literal["causal", "anti-causal"], axis: int
):
    """Compute index adjustment for shifting a time series."""
    adjust_indices = {
        "causal": (0, time_series.shape[axis] - 1),
        "anti-causal": (1, time_series.shape[axis]),
    }
    return adjust_indices[causality]


def shift_time_series(
    time_series: Any,
    predictor_causality: Literal["causal", "anti-causal"] = "causal",
    axis: int = 0,
):
    """Shift time series based on causality of predictor, adding NaNs as needed.

    Shift a time series based on the causality of the predictor and adds NaNs as needed,
    with the operation applied along a specified axis.

    Parameters
    ----------
    time_series :
        The time series to shift, which can be a single array or a pytree of arrays.
        Each array should have a floating-point data type.
    predictor_causality :
        Determines the direction of the shift:
        - 'causal': Shifts the series forward, inserting a NaN at the start.
        - 'anti-causal': Shifts the series backward, appending a NaN at the end.
    axis :
        The axis along which to perform the shift. Must be valid for all arrays in the time series.

    Returns
    -------
    shifted_time_series : Any
        The shifted time series. The structure matches that of `time_series`, with each element
        shifted according to `predictor_causality` and NaNs added accordingly.

    Raises
    ------
    ValueError
        - If `predictor_causality` is not 'causal' or 'anti-causal'.
        - If `axis` is invalid for any array within `time_series`.
        - If any array in `time_series` does not have a floating-point data type.

    Notes
    -----
    The direction of the shift depends on the value of `predictor_causality`:

    - `'causal'`: shift `time_series` one time bin forward and drop final time
      point, so that e.g., `[0, 1, 2]` becomes `[np.nan, 0, 1]`
    - `'anti-causal'`: shift `time_series` one time bin backwards and drop
      first time point, so that e.g., `[0, 1, 2]` becomes `[1, 2, np.nan]`

    """
    # validate axis
    validate_axis(time_series, axis)

    if predictor_causality not in ["causal", "anti-causal"]:
        raise ValueError(
            f"predictor_causality must be one of 'causal', 'anti-causal'. {predictor_causality} provided instead!"
        )

    # compute the start, end indices tree
    adjust_idx = jax.tree_util.tree_map(
        lambda x: _compute_index_adjust(x, predictor_causality, axis), time_series
    )

    # convert to jax ndarray
    time_series = jax.tree_util.tree_map(jnp.asarray, time_series)

    if is_numpy_array_like(time_series):

        if not np.issubdtype(time_series.dtype, np.floating):
            raise ValueError("time_series must have a float dtype!")
        return _pad_dimension(
            jnp.take(time_series, jnp.arange(*adjust_idx), axis=axis),
            axis,
            1,
            predictor_causality,
            jnp.nan,
        )
    else:
        if pytree_map_and_reduce(
            lambda trial: not np.issubdtype(trial.dtype, np.floating), any, time_series
        ):
            raise ValueError("All leaves of time_series must have a float dtype!")
        return jax.tree_util.tree_map(
            lambda trial, idx: _pad_dimension(
                jnp.take(trial, jnp.arange(*idx), axis=axis),
                axis,
                1,
                predictor_causality,
                jnp.nan,
            ),
            time_series,
            adjust_idx,
        )


def row_wise_kron(
    A: jnp.ndarray, C: jnp.ndarray, jit=False, transpose=True
) -> jnp.ndarray:
    r"""Compute the row-wise Kronecker product.

    Compute the row-wise Kronecker product between two matrices using JAX.
    See [\[1\]](#references) for more details on the Kronecker product.

    Parameters
    ----------
    A : jax.numpy.ndarray
        The first matrix.
    C : jax.numpy.ndarray
        The second matrix.
    jit : bool, optional
        Activate Just-in-Time (JIT) compilation. Default is False.
    transpose : bool, optional
        Transpose matrices A and C before computation. Default is True.

    Returns
    -------
    K : jnp.nparray
        The resulting matrix with row-wise Kronecker product.

    Notes
    -----
    This function computes the row-wise Kronecker product between dense matrices A and C
    using JAX for automatic differentiation and GPU acceleration.

    # References
    ------------
    [1] Petersen, Kaare Brandt, and Michael Syskind Pedersen. "The matrix cookbook."
    Technical University of Denmark 7.15 (2008): 510.
    """
    if transpose:
        A = A.T
        C = C.T

    @jax.jit if jit else lambda x: x
    def row_wise_kron(a, c):
        return jnp.kron(a, c)

    K = jax.vmap(row_wise_kron)(A, C)

    if transpose:
        K = K.T

    return K


def assert_has_attribute(obj: Any, attr_name: str):
    """Ensure the object has the given attribute."""
    if not hasattr(obj, attr_name):
        raise AttributeError(
            f"The provided object does not have the required `{attr_name}` attribute!"
        )


def assert_is_callable(func: Callable, func_name: str):
    """Ensure the provided function is callable."""
    if not callable(func):
        raise TypeError(f"The `{func_name}` must be a Callable!")


def assert_returns_ndarray(
    func: Callable, inputs: Union[List[jnp.ndarray], List[float]], func_name: str
):
    """Ensure the function returns a jax.numpy.ndarray."""
    array_out = func(*inputs)
    if not isinstance(array_out, jnp.ndarray):
        raise TypeError(f"The `{func_name}` must return a jax.numpy.ndarray!")


def assert_differentiable(func: Callable, func_name: str):
    """Ensure the function is differentiable."""
    try:
        gradient_fn = jax.grad(func)
        gradient_fn(jnp.array(1.0))
    except Exception as e:
        raise TypeError(f"The `{func_name}` is not differentiable. Error: {str(e)}")


def assert_preserve_shape(
    func: Callable, inputs: List[jnp.ndarray], func_name: str, input_index: int
):
    """Check that the function preserve the input shape."""
    result = func(*inputs)
    if not result.shape == inputs[input_index].shape:
        raise ValueError(f"The `{func_name}` must preserve the input array shape!")


def assert_scalar_func(func: Callable, inputs: List[jnp.ndarray], func_name: str):
    """Check that `func` return an array containing a single scalar."""
    assert_returns_ndarray(func, inputs, func_name)
    array_out = func(*inputs)
    try:
        float(array_out)
    except TypeError:
        raise TypeError(
            f"The `{func_name}` should return a scalar! "
            f"Array of shape {array_out.shape} returned instead!"
        )


def format_repr(
    obj: Base,
    exclude_keys: Optional[List[str]] = None,
    use_name_keys: Optional[List[str]] = None,
    multiline=False,
):
    """
    Format the representation string of an object (`__repr__`).

    This function generates a string representation of an object, including
    the parameters returned by `obj.get_params(deep=False)`. The output excludes
    specified keys and formats certain keys using their class names instead of
    their default string representations.

    Parameters
    ----------
    obj :
        The object inheriting from  Base whose representation string is being formatted.
    exclude_keys :
        List of parameter keys to exclude from the representation. If `None`, no
        keys are excluded. Defaults to `None`.
    use_name_keys :
        List of keys for which the value's `__name__` attribute is used instead
        of the default `__repr__` output. Defaults to an empty list.
    multiline:
        If True, add each parameter on a new line, if False (default), list all parameters
        in a single line.

    Returns
    -------
    :
        A formatted representation string for the object, including its class
        name and relevant parameters.

    Notes
    -----
    - The function includes only non-empty or falsey values (except 0 and False) in the
      representation. The falsey values include empty list or dict, or similar.
    - Parameters are displayed in the order defined in the `obj.__init__`
      method.
    - Parameters with a `shape` attribute (arrays) are excluded from the representation.

    Examples
    --------
    >>> import jax
    >>> from nemos.base_class import Base
    >>> from nemos.regularizer import Ridge
    >>> from nemos.utils import format_repr
    >>> class Example(Base):
    ...     def __init__(self, a, b, c, d):
    ...         self.a = a
    ...         self.b = b
    ...         self.c = c
    ...         self.d = d
    >>> obj = Example(1, jax.numpy.exp, c="hi", d=None)
    >>> format_repr(obj, exclude_keys=["c"], use_name_keys=["b"])
    'Example(a=1, b=exp)'
    """
    exclude_keys = [] if exclude_keys is None else exclude_keys
    use_name_keys = [] if use_name_keys is None else use_name_keys

    init_params = list(inspect.signature(obj.__init__).parameters.keys())
    disp_params = []

    all_params = obj.get_params(deep=False)
    label = all_params.pop("label", None)
    for k, v in all_params.items():
        repr_param = (
            k not in exclude_keys and not hasattr(v, "shape") and (v or v in (0, False))
        )
        if repr_param:
            if k in use_name_keys:
                v = getattr(v, "__name__", repr(v))
            elif isinstance(v, str):
                v = repr(v)
            disp_params.append(f"{k}={v}")
    disp_params = sorted(disp_params, key=lambda x: init_params.index(x.split("=")[0]))
    cls_name = obj.__class__.__name__
    # if label doesn't exist or is the same as the class name (as is the default for
    # basis), then don't use it
    disp_label = (label is not None) and (label != cls_name)
    if multiline:
        if disp_label:
            # cannot use tab directly since IDE converts to spaces, and doctests fail
            tab = "        "
        else:
            tab = "    "
        disp_params = "\n" + tab + f",\n{tab}".join(disp_params) + "\n" + tab[:-4]
        repr_str = (
            f"{repr(label)}:\n    {cls_name}({disp_params})"
            if disp_label
            else f"{cls_name}({disp_params})"
        )
    else:
        disp_params = ", ".join(disp_params)
        repr_str = (
            f"{repr(label)}: {cls_name}({disp_params})"
            if disp_label
            else f"{cls_name}({disp_params})"
        )
    return repr_str


# enable concatenation for pynapple objects.
pynapple_concatenate_jax = support_pynapple(conv_type="jax")(jnp.concatenate)
pynapple_concatenate_numpy = support_pynapple(conv_type="numpy")(np.concatenate)


def _get_terminal_size():
    """Helper to get terminal size for __repr__

    Returns
    -------
    rows:
        Number of rows in the terminal window.
    cols:
        Number of columns in the terminal window.

    """
    cols = 100  # Default
    rows = 2
    try:
        cols, rows = os.get_terminal_size()
    except Exception:
        import shutil

        cols, rows = shutil.get_terminal_size()

    return cols, rows



import joblib
import numpy as np
import pynapple as nap
import pytest
from sklearn import pipeline
from sklearn.model_selection import GridSearchCV

from nemos import basis
from nemos.basis._transformer_basis import TransformerBasis


@pytest.mark.parametrize(
    "bas",
    [
        basis.MSplineEval(5),
        basis.BSplineEval(5),
        basis.CyclicBSplineEval(5),
        basis.OrthExponentialEval(5, decay_rates=np.arange(1, 6)),
        basis.RaisedCosineLinearEval(5),
    ],
)
def test_sklearn_transformer_pipeline(bas, poissonGLM_model_instantiation):
    X, y, model, _, _ = poissonGLM_model_instantiation
    bas = TransformerBasis(bas).set_input_shape(*([1] * bas._n_input_dimensionality))
    pipe = pipeline.Pipeline([("eval", bas), ("fit", model)])

    pipe.fit(X[:, : bas.basis._n_input_dimensionality] ** 2, y)


@pytest.mark.parametrize(
    "bas",
    [
        basis.MSplineEval(5),
        basis.BSplineEval(5),
        basis.CyclicBSplineEval(5),
        basis.RaisedCosineLinearEval(5),
        basis.RaisedCosineLogEval(5),
    ],
)
def test_sklearn_transformer_pipeline_cv(bas, poissonGLM_model_instantiation):
    X, y, model, _, _ = poissonGLM_model_instantiation
    bas = TransformerBasis(bas).set_input_shape(*([1] * bas._n_input_dimensionality))
    pipe = pipeline.Pipeline([("basis", bas), ("fit", model)])
    param_grid = dict(basis__n_basis_funcs=(4, 5, 10))
    gridsearch = GridSearchCV(pipe, param_grid=param_grid, cv=3, error_score="raise")
    gridsearch.fit(X[:, : bas._n_input_dimensionality] ** 2, y)


def test_sklearn_cv_clone(poisson_population_GLM_model):
    X, y, model, _, _ = poisson_population_GLM_model
    bas = basis.CyclicBSplineEval(5)
    bas = TransformerBasis(bas).set_input_shape(*([1] * bas._n_input_dimensionality))
    pipe = pipeline.Pipeline([("basis", bas), ("fit", model)])
    # if feature_mask isn't dropped by the cloning done by gridsearch cv, this will
    # error, because the shape of feature_mask doesn't match the shape of the output of
    # transformer basis with different number of basis funcs
    pipe.fit(X[:, : bas._n_input_dimensionality] ** 2, y)
    param_grid = dict(basis__n_basis_funcs=(4, 8))
    gridsearch = GridSearchCV(pipe, param_grid=param_grid, cv=3, error_score="raise")
    gridsearch.fit(X[:, : bas._n_input_dimensionality] ** 2, y)


@pytest.mark.parametrize(
    "bas",
    [
        basis.MSplineEval(5),
        basis.BSplineEval(5),
        basis.CyclicBSplineEval(5),
        basis.RaisedCosineLinearEval(5),
        basis.RaisedCosineLogEval(5),
    ],
)
def test_sklearn_transformer_pipeline_cv_multiprocess(
    bas, poissonGLM_model_instantiation
):
    X, y, model, _, _ = poissonGLM_model_instantiation
    bas = TransformerBasis(bas).set_input_shape(*([1] * bas._n_input_dimensionality))
    pipe = pipeline.Pipeline([("basis", bas), ("fit", model)])
    param_grid = dict(basis__n_basis_funcs=(4, 5, 10))
    gridsearch = GridSearchCV(
        pipe, param_grid=param_grid, cv=3, n_jobs=3, error_score="raise"
    )
    # use threading instead of fork (this avoids conflicts with jax)
    with joblib.parallel_backend("threading"):
        gridsearch.fit(X[:, : bas._n_input_dimensionality] ** 2, y)


@pytest.mark.parametrize(
    "bas_cls",
    [
        basis.MSplineEval,
        basis.MSplineEval,
        basis.CyclicBSplineEval,
        basis.RaisedCosineLinearEval,
        basis.RaisedCosineLogEval,
    ],
)
def test_sklearn_transformer_pipeline_cv_directly_over_basis(
    bas_cls, poissonGLM_model_instantiation
):
    X, y, model, _, _ = poissonGLM_model_instantiation
    bas = TransformerBasis(bas_cls(5))
    bas.set_input_shape(*([1] * bas._n_input_dimensionality))
    pipe = pipeline.Pipeline([("transformerbasis", bas), ("fit", model)])
    param_grid = dict(
        transformerbasis__basis=(
            bas_cls(5).set_input_shape(*([1] * bas._n_input_dimensionality)),
            bas_cls(10).set_input_shape(*([1] * bas._n_input_dimensionality)),
            bas_cls(20).set_input_shape(*([1] * bas._n_input_dimensionality)),
        )
    )
    gridsearch = GridSearchCV(pipe, param_grid=param_grid, cv=3, error_score="raise")
    gridsearch.fit(X[:, : bas._n_input_dimensionality] ** 2, y)


@pytest.mark.parametrize(
    "bas_cls",
    [
        basis.MSplineEval,
        basis.MSplineEval,
        basis.CyclicBSplineEval,
        basis.RaisedCosineLinearEval,
        basis.RaisedCosineLogEval,
    ],
)
def test_sklearn_transformer_pipeline_cv_illegal_combination(
    bas_cls, poissonGLM_model_instantiation
):
    X, y, model, _, _ = poissonGLM_model_instantiation
    bas = TransformerBasis(bas_cls(5))
    bas.set_input_shape(*([1] * bas._n_input_dimensionality))
    pipe = pipeline.Pipeline([("transformerbasis", bas), ("fit", model)])
    param_grid = dict(
        transformerbasis__basis=(bas_cls(5), bas_cls(10), bas_cls(20)),
        transformerbasis__n_basis_funcs=(4, 5, 10),
    )
    gridsearch = GridSearchCV(pipe, param_grid=param_grid, cv=3, error_score="raise")
    with pytest.raises(
        ValueError,
        match="Set either new basis object or parameters for existing basis, not both.",
    ):
        gridsearch.fit(X[:, : bas._n_input_dimensionality] ** 2, y)


@pytest.mark.parametrize(
    "bas, expected_nans",
    [
        (basis.MSplineEval(5), 0),
        (basis.BSplineEval(5), 0),
        (basis.CyclicBSplineEval(5), 0),
        (basis.OrthExponentialEval(5, decay_rates=np.arange(1, 6)), 0),
        (basis.RaisedCosineLinearEval(5), 0),
        (basis.RaisedCosineLogEval(5), 0),
        (basis.RaisedCosineLogEval(5) + basis.MSplineEval(5), 0),
        (basis.MSplineConv(5, window_size=3), 6),
        (basis.BSplineConv(5, window_size=3), 6),
        (
            basis.CyclicBSplineConv(
                5, window_size=3, conv_kwargs=dict(predictor_causality="acausal")
            ),
            4,
        ),
        (
            basis.OrthExponentialConv(
                5, decay_rates=np.linspace(0.1, 1, 5), window_size=7
            ),
            14,
        ),
        (basis.RaisedCosineLinearConv(5, window_size=3), 6),
        (basis.RaisedCosineLogConv(5, window_size=3), 6),
        (
            basis.RaisedCosineLogConv(5, window_size=3) + basis.MSplineEval(5),
            6,
        ),
        (
            basis.RaisedCosineLogConv(5, window_size=3) * basis.MSplineEval(5),
            6,
        ),
    ],
)
def test_sklearn_transformer_pipeline_pynapple(
    bas, poissonGLM_model_instantiation, expected_nans
):
    X, y, model, _, _ = poissonGLM_model_instantiation

    # transform input to pynapple
    ep = nap.IntervalSet(start=[0, 20.5], end=[20, X.shape[0]])
    X_nap = nap.TsdFrame(t=np.arange(X.shape[0]), d=X, time_support=ep)
    y_nap = nap.Tsd(t=np.arange(X.shape[0]), d=y, time_support=ep)
    bas = TransformerBasis(bas).set_input_shape(*([1] * bas._n_input_dimensionality))

    # fit a pipeline & predict from pynapple
    pipe = pipeline.Pipeline([("eval", bas), ("fit", model)])
    pipe.fit(X_nap[:, : bas.basis._n_input_dimensionality] ** 2, y_nap)

    # get rate
    rate = pipe.predict(X_nap[:, : bas.basis._n_input_dimensionality] ** 2)
    # check rate is Tsd with same time info
    assert isinstance(rate, nap.Tsd)
    assert np.all(rate.t == X_nap.t)
    assert np.all(rate.time_support == X_nap.time_support)
    assert np.sum(np.isnan(rate.d)) == expected_nans



"""Abstract class for regression models."""

# required to get ArrayLike to render correctly
from __future__ import annotations

import abc
import inspect
import warnings
from abc import abstractmethod
from copy import deepcopy
from typing import Any, Dict, NamedTuple, Optional, Tuple, Union

import jax
import jax.numpy as jnp
import jaxopt
from numpy.typing import ArrayLike, NDArray

from . import solvers, utils, validation
from ._regularizer_builder import AVAILABLE_REGULARIZERS, create_regularizer
from .base_class import Base
from .regularizer import Regularizer, UnRegularized
from .typing import DESIGN_INPUT_TYPE, SolverInit, SolverRun, SolverUpdate


class BaseRegressor(Base, abc.ABC):
    """Abstract base class for GLM regression models.

    This class encapsulates the common functionality for Generalized Linear Models (GLM)
    regression models. It provides an abstraction for fitting the model, making predictions,
    scoring the model, simulating responses, and preprocessing data. Concrete classes
    are expected to provide specific implementations of the abstract methods defined here.
    Below is a table listing the default and available solvers for each regularizer.

    | Regularizer   | Default Solver   | Available Solvers                                           |
    | ------------- | ---------------- | ----------------------------------------------------------- |
    | UnRegularized | GradientDescent  | GradientDescent, BFGS, LBFGS, NonlinearCG, ProximalGradient |
    | Ridge         | GradientDescent  | GradientDescent, BFGS, LBFGS, NonlinearCG, ProximalGradient |
    | Lasso         | ProximalGradient | ProximalGradient                                            |
    | GroupLasso    | ProximalGradient | ProximalGradient                                            |

    Parameters
    ----------
    regularizer :
        Regularization to use for model optimization. Defines the regularization scheme
        and related parameters.
        Default is UnRegularized regression.
    regularizer_strength :
        Float that is default None. Sets the regularizer strength. If a user does not pass a value, and it is needed for
        regularization, a warning will be raised and the strength will default to 1.0.
    solver_name :
        Solver to use for model optimization. Defines the optimization scheme and related parameters.
        The solver must be an appropriate match for the chosen regularizer.
        Default is `None`. If no solver specified, one will be chosen based on the regularizer.
        Please see table above for regularizer/optimizer pairings.
    solver_kwargs :
        Optional dictionary for keyword arguments that are passed to the solver when instantiated.
        E.g. stepsize, acceleration, value_and_grad, etc.
         See the jaxopt documentation for details on each solver's kwargs: https://jaxopt.github.io/stable/

    See Also
    --------
    Concrete models:

    - [`GLM`](../glm/#nemos.glm.GLM): A feed-forward GLM implementation.
    - [`PopulationGLM`](../glm/#nemos.glm.PopulationGLM): A population GLM implementation.
    """

    def __init__(
        self,
        regularizer: Union[str, Regularizer] = "UnRegularized",
        regularizer_strength: Optional[float] = None,
        solver_name: str = None,
        solver_kwargs: Optional[dict] = None,
    ):
        self.regularizer = regularizer
        self.regularizer_strength = regularizer_strength

        # no solver name provided, use default
        if solver_name is None:
            self.solver_name = self.regularizer.default_solver
        else:
            self.solver_name = solver_name

        if solver_kwargs is None:
            solver_kwargs = dict()
        self.solver_kwargs = solver_kwargs
        self._solver_init_state = None
        self._solver_update = None
        self._solver_run = None

    @property
    def solver_init_state(self) -> Union[None, SolverInit]:
        """
        Provides the initialization function for the solver's state.

        This function is responsible for initializing the solver's state, necessary for the start
        of the optimization process. It sets up initial values for parameters like gradients and step
        sizes based on the model configuration and input data.

        Returns
        -------
        :
            The function to initialize the state of the solver, if available; otherwise, None if
            the solver has not yet been instantiated.
        """
        return self._solver_init_state

    @property
    def solver_update(self) -> Union[None, SolverUpdate]:
        """
        Provides the function for updating the solver's state during the optimization process.

        This function is used to perform a single update step in the optimization process. It updates
        the model's parameters based on the current state, data, and gradients. It is typically used
        in scenarios where fine-grained control over each optimization step is necessary, such as in
        online learning or complex optimization scenarios.

        Returns
        -------
        :
            The function to update the solver's state, if available; otherwise, None if the solver
            has not yet been instantiated.
        """
        return self._solver_update

    @property
    def solver_run(self) -> Union[None, SolverRun]:
        """
        Provides the function to execute the solver's optimization process.

        This function runs the solver using the initialized parameters and state, performing the
        optimization to fit the model to the data. It iteratively updates the model parameters until
        a stopping criterion is met, such as convergence or exceeding a maximum number of iterations.

        Returns
        -------
        :
            The function to run the solver's optimization process, if available; otherwise, None if
            the solver has not yet been instantiated.
        """
        return self._solver_run

    def set_params(self, **params: Any):
        """Manage warnings in case of multiple parameter settings."""
        # if both regularizer and regularizer_strength are set, then only
        # warn in case the strength is not expected for the regularizer type
        if "regularizer" in params and "regularizer_strength" in params:
            reg = params.pop("regularizer")
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore",
                    category=UserWarning,
                    message="Caution: regularizer strength.*"
                    "|Unused parameter `regularizer_strength`.*",
                )
                super().set_params(regularizer=reg)

        return super().set_params(**params)

    @property
    def regularizer(self) -> Union[None, Regularizer]:
        """Getter for the regularizer attribute."""
        return self._regularizer

    @regularizer.setter
    def regularizer(self, regularizer: Union[str, Regularizer]):
        """Setter for the regularizer attribute."""
        # instantiate regularizer if str
        if isinstance(regularizer, str):
            self._regularizer = create_regularizer(name=regularizer)
        elif isinstance(regularizer, Regularizer):
            self._regularizer = regularizer
        else:
            raise TypeError(
                f"The regularizer should be either a string from "
                f"{AVAILABLE_REGULARIZERS} or an instance of `nemos.regularizer.Regularizer`"
            )

        # force check of regularizer_strength
        # need to use hasattr to avoid class instantiation issues
        if hasattr(self, "_regularizer_strength"):
            self.regularizer_strength = self._regularizer_strength

    @property
    def regularizer_strength(self) -> float:
        """Regularizer strength getter."""
        return self._regularizer_strength

    @regularizer_strength.setter
    def regularizer_strength(self, strength: Union[float, None]):
        # check regularizer strength
        if strength is None and not isinstance(self._regularizer, UnRegularized):
            warnings.warn(
                UserWarning(
                    "Caution: regularizer strength has not been set. Defaulting to 1.0. Please see "
                    "the documentation for best practices in setting regularization strength."
                )
            )
            strength = 1.0
        elif strength is not None:
            try:
                # force conversion to float to prevent weird GPU issues
                strength = float(strength)
            except ValueError:
                # raise a more detailed ValueError
                raise ValueError(
                    f"Could not convert the regularizer strength: {strength} to a float."
                )
            if isinstance(self._regularizer, UnRegularized):
                warnings.warn(
                    UserWarning(
                        "Unused parameter `regularizer_strength` for UnRegularized GLM. "
                        "The regularizer strength parameter is not required and won't be used when the regularizer "
                        "is set to UnRegularized."
                    )
                )

        self._regularizer_strength = strength

    @property
    def solver_name(self) -> str:
        """Getter for the solver_name attribute."""
        return self._solver_name

    @solver_name.setter
    def solver_name(self, solver_name: str):
        """Setter for the solver_name attribute."""
        # check if solver str passed is valid for regularizer
        if solver_name not in self._regularizer.allowed_solvers:
            raise ValueError(
                f"The solver: {solver_name} is not allowed for "
                f"{self._regularizer.__class__.__name__} regularization. Allowed solvers are "
                f"{self._regularizer.allowed_solvers}."
            )
        self._solver_name = solver_name

    @property
    def solver_kwargs(self):
        """Getter for the solver_kwargs attribute."""
        return self._solver_kwargs

    @solver_kwargs.setter
    def solver_kwargs(self, solver_kwargs: dict):
        """Setter for the solver_kwargs attribute."""
        self._check_solver_kwargs(
            self._get_solver_class(self.solver_name), solver_kwargs
        )
        self._solver_kwargs = solver_kwargs

    @staticmethod
    def _check_solver_kwargs(solver_class, solver_kwargs):
        """
        Check if provided solver keyword arguments are valid.

        Parameters
        ----------
        solver_class :
            Class of the solver.
        solver_kwargs :
            Additional keyword arguments for the solver.

        Raises
        ------
        NameError
            If any of the solver keyword arguments are not valid.
        """
        solver_args = inspect.getfullargspec(solver_class).args
        undefined_kwargs = set(solver_kwargs.keys()).difference(solver_args)
        if undefined_kwargs:
            raise NameError(
                f"kwargs {undefined_kwargs} in solver_kwargs not a kwarg for {solver_class.__name__}!"
            )

    def instantiate_solver(
        self, *args, solver_kwargs: Optional[dict] = None
    ) -> BaseRegressor:
        """
        Instantiate the solver with the provided loss function.

        Instantiate the solver with the provided loss function, and store callable functions
        that initialize the solver state, update the model parameters, and run the optimization
        as attributes.

        This method creates a solver instance from nemos.solvers or the jaxopt library, tailored to
        the specific loss function and regularization approach defined by the Regularizer instance.
        It also handles the proximal operator if required for the optimization method. The returned
        functions are directly usable in optimization loops, simplifying the syntax by pre-setting
        common arguments like regularization strength and other hyperparameters.

        Parameters
        ----------
        *args:
            Positional arguments for the jaxopt `solver.run` method, e.g. the regularizing
            strength for proximal gradient methods.
        solver_kwargs:
            Optional dictionary with the solver kwargs.
            If nothing is provided, it defaults to self.solver_kwargs.

        Returns
        -------
        :
            The instance itself for method chaining.
        """
        # final check that solver is valid for chosen regularizer
        if self.solver_name not in self.regularizer.allowed_solvers:
            raise ValueError(
                f"The solver: {self.solver_name} is not allowed for "
                f"{self._regularizer.__class__.__name__} regularization. Allowed solvers are "
                f"{self._regularizer.allowed_solvers}."
            )

        # only use penalized loss if not using proximal gradient descent
        # In proximal method you must use the unpenalized loss independently
        # of what regularizer you are using.
        if self.solver_name not in ("ProximalGradient", "ProxSVRG"):
            loss = self.regularizer.penalized_loss(
                self._predict_and_compute_loss, self.regularizer_strength
            )
        else:
            loss = self._predict_and_compute_loss

        if solver_kwargs is None:
            # copy dictionary of kwargs to avoid modifying user settings
            solver_kwargs = deepcopy(self.solver_kwargs)

        # check that the loss is Callable
        utils.assert_is_callable(loss, "loss")

        # some parsing to make sure solver gets instantiated properly
        if self.solver_name in ("ProximalGradient", "ProxSVRG"):
            if "prox" in self.solver_kwargs:
                raise ValueError(
                    "Proximal operator specification is not permitted. "
                    "The proximal operator is automatically determined based on the selected regularizer. "
                    "Please remove the 'prox' argument from the `solver_kwargs` "
                )

            solver_kwargs.update(prox=self.regularizer.get_proximal_operator())
            # add self.regularizer_strength to args
            args += (self.regularizer_strength,)

        (
            solver_run_kwargs,
            solver_init_state_kwargs,
            solver_update_kwargs,
            solver_init_kwargs,
        ) = self._inspect_solver_kwargs(solver_kwargs)

        # instantiate the solver
        solver = self._get_solver_class(self.solver_name)(
            fun=loss, **solver_init_kwargs
        )

        self._solver_loss_fun_ = loss

        def solver_run(
            init_params: Tuple[DESIGN_INPUT_TYPE, jnp.ndarray], *run_args: jnp.ndarray
        ) -> jaxopt.OptStep:
            return solver.run(init_params, *args, *run_args, **solver_run_kwargs)

        def solver_update(params, state, *run_args, **run_kwargs) -> jaxopt.OptStep:
            return solver.update(
                params, state, *args, *run_args, **solver_update_kwargs, **run_kwargs
            )

        def solver_init_state(params, *run_args, **run_kwargs) -> NamedTuple:
            return solver.init_state(
                params,
                *run_args,
                **run_kwargs,
                **solver_init_state_kwargs,
            )

        self._solver_init_state = solver_init_state
        self._solver_update = solver_update
        self._solver_run = solver_run
        return self

    def _inspect_solver_kwargs(
        self, solver_kwargs: dict
    ) -> Tuple[Dict[str, Any], Dict[str, Any], Dict[str, Any], Dict[str, Any]]:
        """Inspect and categorize the solver keyword arguments.

        This method inspects the provided `solver_kwargs` dictionary and categorizes
        the keyword arguments based on which solver functions they apply to:
        `run`, `init_state`, `update`, and `__init__`. This ensures that the
        appropriate arguments are passed to each function when the solver is used.

        Parameters
        ----------
        solver_kwargs :
            Dictionary containing keyword arguments for the solver.

        Returns
        -------
        :
            A tuple containing four dictionaries:
            - solver_run_kwargs: Arguments for the solver's `run` method.
            - solver_init_state_kwargs: Arguments for the solver's `init_state` method.
            - solver_update_kwargs: Arguments for the solver's `update` method.
            - solver_init_kwargs: Arguments for the solver's `__init__` constructor.
        """
        solver_run_kwargs = dict()
        solver_init_state_kwargs = dict()
        solver_update_kwargs = dict()
        solver_init_kwargs = dict()

        if solver_kwargs:
            # instantiate a solver to then inspect the params of its various functions
            solver = self._get_solver_class(self.solver_name)

            for key, value in solver_kwargs.items():
                if key in inspect.getfullargspec(solver.run).args:
                    solver_run_kwargs[key] = value
                if key in inspect.getfullargspec(solver.init_state).args:
                    solver_init_state_kwargs[key] = value
                if key in inspect.getfullargspec(solver.update).args:
                    solver_update_kwargs[key] = value
                if key in inspect.getfullargspec(solver.__init__).args:
                    solver_init_kwargs[key] = value

        return (
            solver_run_kwargs,
            solver_init_state_kwargs,
            solver_update_kwargs,
            solver_init_kwargs,
        )

    @abc.abstractmethod
    def fit(self, X: DESIGN_INPUT_TYPE, y: Union[NDArray, jnp.ndarray]):
        """Fit the model to neural activity."""
        pass

    @abc.abstractmethod
    def predict(self, X: DESIGN_INPUT_TYPE) -> jnp.ndarray:
        """Predict rates based on fit parameters."""
        pass

    @abc.abstractmethod
    def score(
        self,
        X: DESIGN_INPUT_TYPE,
        y: Union[NDArray, jnp.ndarray],
        # may include score_type or other additional model dependent kwargs
        **kwargs,
    ) -> jnp.ndarray:
        """Score the predicted firing rates (based on fit) to the target neural activity."""
        pass

    @abc.abstractmethod
    def simulate(
        self,
        random_key: jax.Array,
        feed_forward_input: DESIGN_INPUT_TYPE,
    ):
        """Simulate neural activity in response to a feed-forward input and recurrent activity."""
        pass

    @staticmethod
    @abc.abstractmethod
    def _check_params(
        params: Tuple[Union[DESIGN_INPUT_TYPE, ArrayLike], ArrayLike],
        data_type: Optional[jnp.dtype] = None,
    ) -> Tuple[DESIGN_INPUT_TYPE, jnp.ndarray]:
        """
        Validate the dimensions and consistency of parameters.

        This function checks the consistency of shapes and dimensions for model
        parameters.
        It ensures that the parameters and data are compatible for the model.

        """
        pass

    @staticmethod
    @abc.abstractmethod
    def _check_input_dimensionality(
        X: Optional[Union[DESIGN_INPUT_TYPE, jnp.ndarray]] = None,
        y: Optional[jnp.ndarray] = None,
    ):
        pass

    @abc.abstractmethod
    def _get_coef_and_intercept(self) -> Tuple[Any, Any]:
        """Pack coef_ and intercept_  into a params pytree."""
        pass

    @abc.abstractmethod
    def _set_coef_and_intercept(self, params: Any):
        """Unpack and store params pytree to coef_ and intercept_."""
        pass

    @staticmethod
    @abc.abstractmethod
    def _check_input_and_params_consistency(
        params: Tuple[Union[DESIGN_INPUT_TYPE, jnp.ndarray], jnp.ndarray],
        X: Optional[Union[DESIGN_INPUT_TYPE, jnp.ndarray]] = None,
        y: Optional[jnp.ndarray] = None,
    ):
        """Validate the number of features in model parameters and input arguments.

        Raises
        ------
        ValueError
            - if the number of features is inconsistent between params[1] and X
              (when provided).

        """
        pass

    @staticmethod
    def _check_input_n_timepoints(
        X: Union[DESIGN_INPUT_TYPE, jnp.ndarray], y: jnp.ndarray
    ):
        if y.shape[0] != X.shape[0]:
            raise ValueError(
                "The number of time-points in X and y must agree. "
                f"X has {X.shape[0]} time-points, "
                f"y has {y.shape[0]} instead!"
            )

    @abc.abstractmethod
    def _predict_and_compute_loss(self, params, X, y):
        """Loss function for a given model to be optimized over."""
        pass

    def _validate(
        self,
        X: Union[DESIGN_INPUT_TYPE, jnp.ndarray],
        y: Union[NDArray, jnp.ndarray],
        init_params: Tuple[DESIGN_INPUT_TYPE, jnp.ndarray],
    ):
        # check input dimensionality
        self._check_input_dimensionality(X, y)
        self._check_input_n_timepoints(X, y)

        # error if all samples are invalid
        validation.error_all_invalid(X, y)

        # validate input and params consistency
        init_params = self._check_params(init_params)

        # validate input and params consistency
        self._check_input_and_params_consistency(init_params, X=X, y=y)

    @abc.abstractmethod
    def update(
        self,
        params: Tuple[jnp.ndarray, jnp.ndarray],
        opt_state: NamedTuple,
        X: DESIGN_INPUT_TYPE,
        y: jnp.ndarray,
        *args,
        **kwargs,
    ) -> jaxopt.OptStep:
        """Run a single update step of the jaxopt solver."""
        pass

    @abc.abstractmethod
    def initialize_params(
        self,
        X: DESIGN_INPUT_TYPE,
        y: jnp.ndarray,
        params: Optional = None,
    ) -> Union[Any, NamedTuple]:
        """Initialize the solver's state and optionally sets initial model parameters for the optimization."""
        pass

    @abc.abstractmethod
    def initialize_state(
        self,
        X: DESIGN_INPUT_TYPE,
        y: jnp.ndarray,
        init_params,
    ) -> Union[Any, NamedTuple]:
        """Initialize the state of the solver for running fit and update."""
        pass

    @staticmethod
    def _get_solver_class(solver_name: str):
        """
        Find a solver class first looking in nemos.solvers, then in jaxopt.

        Parameters
        ----------
        solver_name : str
            Name of the solver class to load.

        Returns
        -------
        solver_class :
            Solver class ready to be instantiated.

        Raises
        ------
        AttributeError
            If a solver class with that name is not found.
        """
        try:
            solver_class = getattr(solvers, solver_name)
        except AttributeError:
            try:
                solver_class = getattr(jaxopt, solver_name)
            except AttributeError:
                raise AttributeError(
                    f"Could not find {solver_name} in nemos.solvers or jaxopt"
                )

        return solver_class

    def _optimize_solver_params(self, X: DESIGN_INPUT_TYPE, y: jnp.ndarray) -> dict:
        """
        Compute and update solver parameters with optimal defaults if available.

        This method checks the current solver configuration and, if an optimal
        configuration is known for the given model parameters, computes the optimal
        batch size, step size, and other hyperparameters to ensure faster convergence.

        Parameters
        ----------
        X :
            Input data used to compute smoothness and strong convexity constants.
        y :
            Target values used in conjunction with X for the same purpose.

        Returns
        -------
        :
            A dictionary containing the solver parameters, updated with optimal defaults
            where applicable.

        """
        # Start with a copy of the existing solver parameters
        new_solver_kwargs = self.solver_kwargs.copy()

        # get the model specific configs
        compute_defaults, compute_l_smooth, strong_convexity = (
            self._get_optimal_solver_params_config()
        )
        if compute_defaults and compute_l_smooth:
            # Check if the user has provided batch size or stepsize, or else use None
            batch_size = new_solver_kwargs.get("batch_size", None)
            stepsize = new_solver_kwargs.get("stepsize", None)

            # Compute the optimal batch size and stepsize based on smoothness, strong convexity, etc.
            new_params = compute_defaults(
                compute_l_smooth,
                X,
                y,
                batch_size=batch_size,
                stepsize=stepsize,
                strong_convexity=strong_convexity,
            )

            # Update the solver parameters with the computed optimal values
            new_solver_kwargs.update(new_params)

        return new_solver_kwargs

    @abstractmethod
    def _get_optimal_solver_params_config(self):
        """Return the functions for computing default step and batch size for the solver."""
        pass
